#!/bin/bash

# Design:
# Executes main function with different ip_block_full_{add/del/get}/commit
# functions. For each router that needs to be synced with the ip_block.load
# you should create separate hook and put its full path in file:
#
# 	/cfg/userinfo/hooks/events.sh.hooks
#
# You don't need to invent any logic on WHAT and WHEN to send to routers. You only
# should define HOW to send it in hook. Few add/del and single get/commit are
# called here. You can do some batch-style code to make add/del faster
# for example, you can (it's little bit complicated) generate commands-file that
# will be sent in commit function.
#
# For quagga blackhole please use /app/bgp_blackhole instead of /app/reductor.
# IP lists are synced between apps, everything should work out of box.
# If you still use Carbon Reductor 7 consider migration to Carbon Reductor 8.

set -eu
. /usr/local/Reductor/etc/const
# shellcheck disable=SC1090
. "$CONFIG"

# please redefine them in hook:
# CR7: http://docs.carbonsoft.ru/55934984
# CR8: http://docs.carbonsoft.ru/68354281

LOCKDIR=/var/lock/reductor/
LOCKFILE_IP="$LOCKDIR/ip.lock"
FD_IP=5
LOCKFILE_EVENTS="$LOCKDIR/events.sh.lock"
FD_EVENTS=6
EVENT_DIR="$TMPDIR/events.sh/$$/"
EVENT_FILE=${EVENT_DIR}/ip_block
CUSTOM_EXIT=1
trap __exit EXIT

__custom_exit() {
	release_lock "$LOCKFILE_IP" "$FD_IP" &>/dev/null || true
	release_lock "$LOCKFILE_EVENTS" "$FD_EVENTS" &>/dev/null || true
	log "Подчищаем временную директорию $EVENT_DIR"
	rm -rf "$EVENT_DIR"
}

prepare() {
	mkdir -p "$TMPDIR" "$LOCKDIR" "${DATADIR}" "$EVENT_DIR"
	set_lock "$LOCKFILE_EVENTS" 60 "$FD_EVENTS"
	set_lock "$LOCKFILE_IP" 60 "$FD_IP"
	local IP_BLOCK_FULL_LIST_DEFAULT="$LISTDIR/load/ip_block.load"
	local IP_BLOCK_FULL_LIST
	if [ "${filter['events_subnet']:-0}" = '1' ]; then
		IP_BLOCK_FULL_LIST_DEFAULT="$LISTDIR/load/ip_subnet_block.load"
	fi
	IP_BLOCK_FULL_LIST="${IP_BLOCK_FULL_LIST:-$IP_BLOCK_FULL_LIST_DEFAULT}"
	log "Синхронизируем с маршрутизатором список IP: $IP_BLOCK_FULL_LIST"
	cp -a "$IP_BLOCK_FULL_LIST" "$EVENT_FILE"
	release_lock "$LOCKFILE_IP" "$FD_IP"

}

use_hook() {
	local hook="$1"
	# shellcheck disable=SC1090
	[ -x "$hook" ] && . "$hook"
}

diff_data_only() {
	# shellcheck disable=SC2196
	egrep '^(\+|\-)' | egrep -v '^(\+|\-){3}'
}


__maskify() {
	sed -E "s|^$ip_regex$|&/32|g"
}

maskify() {
	if [ "${filter['events_subnet']:-0}" = '0' ]; then
		cat
	else
		__maskify
	fi
}

diff_to_events() {
	sed -e 's/-/ip_block_full_add /; s/+/ip_block_full_del /'
}

files_to_events() {
	local old="$1"
	local new="$2"
	sort_lists "$old" "$new"
	# sort -r is for del before add. In other cases subnet logic may be bugged
	diff -U 0 "$new" "$old" | diff_data_only | diff_to_events | sort -r
}

sort_lists() {
	for list in "$@"; do
		touch "$list"
		${BINDIR}/sort.sh "$list"
	done
}

process_events() {
	local rc
	local events=0
	while read event ip _; do
		rc=0
		((events++)) || true
		"$event" "$ip" || rc=$?
		log "$event $ip (event #$events rc=$rc)"
	done
	# we don't need to commit nothing
	if [ "$events" = '0' ]; then
		log "Всё и так уже синхронизировано."
	else
		commit
	fi
}

mikrotik_retry() {
	local output="$1"
	[ ! -s "$output" ] || return 0
	log "Не получен (либо пуст) список IP, пробуем запросить заново через 1 секунду."
	sleep 1
	ip_block_full_get | sort -u > "$output"
}

__main() {
	log "tick begin"
	prepare
	local FILE="$DATADIR/ip_block_full"
	local NEW="$FILE.new"
	local OLD="$FILE.old"
	ip_block_full_get | sort -u | maskify > "$OLD"
	mikrotik_retry "$OLD"
	maskify < "$EVENT_FILE" > "$NEW"
	log "На сервере содержится $(wc -l < "$OLD") записей"
	log "Для полной блокировки имеется $(wc -l < "$NEW") записей"
	files_to_events "$OLD" "$NEW" | process_events
	log "tick end"
}

main() {
	local hooklistfile="$HOOKDIR/events.sh.hooks"
	local hook hooks
	declare -a hooks
	hooks=()
	if [ -s "$hooklistfile" ]; then
		while read line; do
			hooks=( "${hooks[@]:-}" "$line" )
		done < "$hooklistfile"
	fi
	# backward compatibility
	hooks=( "${hooks[@]:-}" "$HOOKDIR/${0##*/}" )
	for hook in "${hooks[@]}"; do
		[ -x "$hook" ] || continue
		log "Синхронизация с помощью $hook"

		# Обнуляем функции от предыдущего хука.
		ip_block_full_add() { return 0; }
		ip_block_full_del() { return 0; }
		ip_block_full_get() { return 0; }
		commit() { return 0; }
		unset IP_BLOCK_FULL_LIST

		# shellcheck disable=SC1090
		. "$hook"
		__main "$@"
	done
	release_lock "$LOCKFILE_EVENTS" "$FD_EVENTS"
	exit 0
}

main "$@"
