#!/bin/bash

. /etc/init.d/functions
. /usr/local/Reductor/etc/const

# shellcheck disable=SC1090
. "$CONFIG"

set -eu

TCP_REJECT="REJECT --reject-with tcp-reset"
readonly DUAL_TCP_REJECT="DUALRST --dualrst-with tcp-reset --rst --both"
if ${BINDIR}/dnsenabled --dualrst; then
	TCP_REJECT="$DUAL_TCP_REJECT"
fi

if [ -f "$LOCKDIR/maintain.lock" ]; then
	readonly MAINTAINANCE_REQUIRED='1' # блокировка до окончания загрузки URL
elif ${BINDIR}/dnsenabled --revisor; then
	readonly MAINTAINANCE_REQUIRED='2' # блокировка на время перестроения iptables
else
	readonly MAINTAINANCE_REQUIRED='0' # вообще никакой блокировки не надо
fi

log "MAINTAINANCE_REQUIRED = $MAINTAINANCE_REQUIRED"

read -a revisor_ips <<< "${filter['revisor_ip']:-}"
readonly PREFIX='reductor_'
readonly FORWARD="${PREFIX}forward"
readonly INPUT="${PREFIX}input"
readonly POSTROUTING="${PREFIX}postrouting"
readonly PREROUTING="${PREFIX}prerouting"

create_and_flush_chain() {
	chain="$1"
	table="${2:-filter}"
	iptables -t "$table" -N "$chain" 2>/dev/null || true
	iptables -t "$table" -F "$chain"
}

add_skip_check_rule() {
	if [ -s "$HOOKDIR/skip_check.list" ]; then
		log  "Добавляется правило DROP для белого списка абонентов"
		iptables -A "$FORWARD" -m set --set skip_check src -j DROP
	fi
}

add_custom_rules_http() {
	local blacklist_id ipset_id redirect_url comment
	if [ -s "$HOOKDIR/custom_bindings" ]; then
		local FORBIDDEN_TARGET="$1"
		log "Добавляются дополнительные правила фильтрации http"
		while read blacklist_id ipset_id redirect_url comment; do
			if [ "$ipset_id" = '-1' ] && [ "$blacklist_id" = '0' ]; then
				log "Удаляем дефолтное фильтрующее правило из-за записи в $HOOKDIR/custom_bindings: $blacklist_id $ipset_id $redirect_url $comment"
				iptables -A ${PREFIX}http -p tcp -m reductor --id "$blacklist_id" --comment "$comment" -j $FORBIDDEN_TARGET $redirect_url
				iptables -D ${PREFIX}http 1
				continue
			fi
			iptables -A ${PREFIX}http -p tcp -m set --match-set "custom_$ipset_id" src -m reductor --id $blacklist_id --comment "$comment" -j $FORBIDDEN_TARGET $redirect_url
		done < "$HOOKDIR/custom_bindings"
	fi
}

add_custom_rules_dns() {
	local bindings="$HOOKDIR/custom_bindings_dns"
	local list_id ipset_id redirect_ip exact_mode blackwhite comment
	if [ ! -s "$bindings" ]; then
		return 0
	fi
	while read list_id ipset_id redirect_ip exact_mode blackwhite comment; do
		if [ "$exact_mode" = 'exact' ]; then
			exact_mode=' --exact'
		elif [ "$exact_mode" = 'with_subdomains' ]; then
			exact_mode=''
		else
			log "WARNING: не указан exact_mode для $list_id $ipset_id $redirect_ip $exact_mode $comment"
			continue
		fi
		if [ "$blackwhite" != 'black' ] && [ "$blackwhite" != 'white' ]; then
			log "WARNING: неизвестно, является ли список $list_id $ipset_id $redirect_ip $exact_mode $comment чёрным или белым"
			continue
		fi
		# shellcheck disable=SC2154
		if [ "$blackwhite" = 'black' ] && ! echo "$redirect_ip" | egrep -q "$ip_regex"; then
			log "WARNING: IP редиректа $redirect_ip не является IP адресом ($comment)"
			continue
		fi
		if [ "$ipset_id" = '-' ]; then
			ipset_id=''
		else
			ipset_id="-m set --match-set custom_$ipset_id src"
		fi
		if [ "$blackwhite" = 'black' ]; then
			iptables -A dns -p udp $ipset_id -m dnsmatch --id "$list_id" $exact_mode --comment "$comment" -j DNSREPLY --dnsreply-ip "$redirect_ip"
		else
			iptables -I dns -p udp $ipset_id -m dnsmatch --id "$list_id" $exact_mode --comment "$comment" -j RETURN
		fi
	done < "$bindings"
}

add_custom_rules_https() {
	if [ -s "$HOOKDIR/custom_bindings_https" ]; then
		local TARGET="$TCP_REJECT"
		log "Добавляются дополнительные правила фильтрации https"
		while read src dst _; do
			iptables -A ${PREFIX}https_dst -p tcp -m set --match-set "custom_$src" src -m set --match-set "custom_$dst" dst -j $TARGET
			iptables -A ${PREFIX}https_src -p tcp -m set --match-set "custom_$src" src -m set --match-set "custom_$dst" dst -j $TARGET
		done < "$HOOKDIR/custom_bindings_https"
	fi
}

admin_notify() {
	local is_readonly
	local page="http://deny.carbonsoft.ru/whereismymoney"
	is_readonly="$(python -m json.tool /var/reg/CRB-Reductor.key | grep '"cf_regreadonly":' | egrep -o [0-9])" || true
	# shellcheck disable=SC2154
	if [ "${misc['admin_notify']:-0}" = '1' ] && [ "$is_readonly" = '1' ]; then
		log "Добавляем напоминание для администраторов о скорой блокировке"
		if [ -d /cfg ]; then
			iptables -I ${PREFIX}http ! -d "$CARBONSOFT_BLACKHOLE" -p tcp --tcp-flags PSH PSH -m set --set base_auth_users src -m limit --limit 5/hour -j FORBIDDEN --forbidden-url "$page" || true
			iptables -I ${PREFIX}http ! -d "$CARBONSOFT_BLACKHOLE" -p tcp --tcp-flags PSH PSH -m set --set base_admins src -m limit --limit 5/hour -j FORBIDDEN --forbidden-url "$page" || true
		else
			iptables -I ${PREFIX}http ! -d "$CARBONSOFT_BLACKHOLE" -p tcp --tcp-flags PSH PSH -m set --set reductor_admins src -m limit --limit 5/hour -j FORBIDDEN --forbidden-url "$page"
		fi
	fi
}

http() {
	log "Настраиваем фильтрацию http"
	create_and_flush_chain "${PREFIX}http"
	local TARGET="$TCP_REJECT"
	local FORBIDDEN_TARGET="FORBIDDEN --forbidden-url "
	if [ "${filter['forbidden']}" = '1' ]; then
		local url="${filter['forbidden_url']:-http://helpdesk.carbonsoft.ru/deny}"
		TARGET="FORBIDDEN"
		[ "${filter['save_domain']:-0}" == '1' ] && TARGET="$TARGET --save-domain"
		[ "${filter['forbidden_log']:-0}" == '1' ] && TARGET="$TARGET --log"
		FORBIDDEN_TARGET="$TARGET --forbidden-url" #for use in custom redirect
		TARGET="$FORBIDDEN_TARGET $url"
	fi
	local traffic_to_filter="-p tcp -m set --match-set port_http dst"
	if [ -s "$LISTDIR/provider/our.white_http_iplist" ]; then
		log "Добавляем белый список ip (http)"
		iptables -A ${PREFIX}http -m set --match-set ip_block_whitelist dst -j DROP
	fi
	iptables -A ${PREFIX}http -p tcp -m reductor --id 0 --comment 'rkn' -j $TARGET
	admin_notify
	add_custom_rules_http "$FORBIDDEN_TARGET"
	iptables -A "$FORWARD" $traffic_to_filter -j ${PREFIX}http
}

sni() {
	local nginx_ip
	[ "${filter['sni']:-0}" = '1' ] || return 0
	log "Настраиваем фильтрацию HTTPS по SNI"
	if ! ${BINDIR}/dnsenabled; then
		iptables -A ${PREFIX}https_dst -m length --length 61: -p tcp -m snimatch --id 0 --comment 'sni https?' -j $TCP_REJECT
		return
	fi
	if nginx_ip="$(${BINDIR}/dnsenabled --nginx-ip)"; then
		iptables -A ${PREFIX}https_dst -d ${nginx_ip} -j DROP
	fi
	if ${BINDIR}/dnsenabled --proxy-may-be-used; then
		#TODO: proxy can be installed on the same machine
		iptables -A ${PREFIX}https_dst -d ${filter['dns_proxy_ip']} -j DROP
	fi
	iptables -A ${PREFIX}https_dst -m length --length 61: -p tcp -m snimatch --id 0 --comment 'sni https?' -j $TCP_REJECT
}

ip_https() {
	local ip
	for direct in dst src; do
		for ipset in ip_https ip_https_plus; do
			if [ "${filter['https_by_ip']}" = '0' ] && [ "$ipset" = 'ip_https_plus' ]; then
				log "Блокируем HTTPS по IP только в крайнем случае."
				if ${BINDIR}/dnsenabled --revisor && [ "$direct" == 'dst' ]; then
					log "Но ревизору блокируем его всегда."
					for ip in "${revisor_ips[@]}"; do
						iptables -A "${PREFIX}https_$direct" -p tcp -s "$ip" -m set --match-set $ipset "$direct" -j $TCP_REJECT
					done
				fi
				continue
			fi
			iptables -A "${PREFIX}https_$direct" -p tcp -m set --match-set "$ipset" "$direct" -j $TCP_REJECT
		done
	done
}

ip_https_whitelist() {
	if [ -f "$LISTDIR/provider/our.ip_https_whitelist" ]; then
		log "Добавляем белый список ip (https)"
		iptables -A ${PREFIX}https_dst -m set --match-set ip_https_whitelist dst -j DROP
		iptables -A ${PREFIX}https_src -m set --match-set ip_https_whitelist src -j DROP
	fi
}

https() {
	log "Настраиваем фильтрацию https"
	create_and_flush_chain "${PREFIX}https_src"
	create_and_flush_chain "${PREFIX}https_dst"
	ip_https_whitelist
	# sni после ip_https т.к. незачем напрягать sni, если это всё равно отреджектится по ip
	ip_https
	sni
	add_custom_rules_https
	iptables -A "$FORWARD" -p tcp -m set --match-set port_https dst -j ${PREFIX}https_dst
	iptables -A "$FORWARD" -p tcp -m set --match-set port_https src -j ${PREFIX}https_src
}

ip_port() {
	if [ "${filter['cardsharing']:-}" = '1' ]; then
		log "Настраиваем фильтрацию кардшаринга и прочего"
		create_and_flush_chain "${PREFIX}other_traffic"
		for side in src dst; do
			iptables -A ${PREFIX}other_traffic -p tcp -m set --match-set ip_tcp_udp "$side" -m set --match-set port_tcp "$side" -j $TCP_REJECT
			iptables -A ${PREFIX}other_traffic -p udp -m set --match-set ip_tcp_udp "$side" -m set --match-set port_udp "$side" -j REJECT --reject-with icmp-port-unreachable
		done
		iptables -A "$FORWARD" -j ${PREFIX}other_traffic
	fi
}

ip_block() {
	log "Настраиваем фильтрацию IP адресов"
	create_and_flush_chain "${PREFIX}ip_block"
	iptables -A ${PREFIX}ip_block -p tcp -j $TCP_REJECT
	iptables -A ${PREFIX}ip_block -p udp -j REJECT --reject-with icmp-port-unreachable
	iptables -A ${PREFIX}ip_block -j REJECT --reject-with icmp-proto-unreachable
	iptables -A "$FORWARD" -m set --match-set ip_block dst,dst -j ${PREFIX}ip_block
}

filter_policy() {
	iptables -P INPUT ACCEPT
	iptables -P FORWARD DROP
	iptables -P OUTPUT ACCEPT
}

dns_post_revisor_hook() {
	return 0
}

dns() {
	$BINDIR/dnsenabled || return 0
	local IP
	local icase=""
	IP="$(${BINDIR}/dnsenabled --nginx-ip)"
	log "Настраиваем фильтрацию DNS"
	if [ "$IP" != "${IP%%:*}" ]; then
		$SETCOLOR_FAILURE
		log "WARNING: обрезаем вероятно порт, указанный рядом с IP, DNS ничего не знает о портах"
		$SETCOLOR_NORMAL
		IP="${IP%%:*}"
	fi
	create_and_flush_chain "${PREFIX}dns"
	if ${BINDIR}/dnsenabled --revisor; then
		local ip
		log "Добавляем правила icase-спуфинга для ревизора"
		create_and_flush_chain "${PREFIX}dns_revisor"
		for ip in "${revisor_ips[@]}"; do
			iptables -I "${PREFIX}dns" -s "$ip" -j "${PREFIX}dns_revisor"
		done
		dns_post_revisor_hook subchain
	else
		$SETCOLOR_FAILURE
		log "Правило для icase-спуфинга будет использоваться для всех, это повышает нагрузку на сервер, укажите IP ревизора."
		$SETCOLOR_NORMAL
		icase=" --icase"
	fi

	local MASK_MATCH="-p udp -m dnsmatch --id 0 --comment domain_mask $icase"
	local MASK_TARGET="DNSREPLY --dnsreply-ip $IP"
	iptables -A ${PREFIX}dns $MASK_MATCH -j $MASK_TARGET
	if ${BINDIR}/dnsenabled --revisor; then
		iptables -A ${PREFIX}dns_revisor $MASK_MATCH --icase -j $MASK_TARGET
	fi

	if ${BINDIR}/dnsenabled --separate-exact-and-mask; then
		local EXACT_MATCH="-p udp -m dnsmatch --id 1000 --exact --comment domain_exact $icase"
		local EXACT_TARGET="DNSREPLY --dnsreply-ip $IP"
		iptables -A "${PREFIX}dns" $EXACT_MATCH -j $EXACT_TARGET
		if ${BINDIR}/dnsenabled --revisor; then
			iptables -A "${PREFIX}dns_revisor" $EXACT_MATCH --icase -j $EXACT_TARGET
		fi
	fi

	if ${BINDIR}/dnsenabled --proxy-https; then
		local PROXY_HTTPS_MATCH="! -s ${filter['dns_proxy_ip']} -p udp -m dnsmatch --id 2000 --comment domain_proxy $icase"
		local PROXY_HTTPS_TARGET="DNSREPLY --dnsreply-ip ${filter['dns_proxy_ip']}"
		iptables -A ${PREFIX}dns $PROXY_HTTPS_MATCH -j $PROXY_HTTPS_TARGET
		if ${BINDIR}/dnsenabled --revisor; then
			iptables -A "${PREFIX}dns_revisor" $PROXY_HTTPS_MATCH --icase -j $PROXY_HTTPS_TARGET
		fi
	fi

	if ${BINDIR}/dnsenabled --proxy-hsts; then
		local PROXY_HSTS_MATCH="! -s ${filter['dns_proxy_ip']} -p udp -m dnsmatch --id 3000 --comment domain_proxy $icase"
		local PROXY_HSTS_TARGET="DNSREPLY --dnsreply-ip ${filter['dns_proxy_ip']}"
		iptables -A "${PREFIX}dns" $PROXY_HSTS_MATCH -j $PROXY_HSTS_TARGET
		if ${BINDIR}/dnsenabled --revisor; then
			iptables -A "${PREFIX}dns_revisor" $PROXY_HSTS_MATCH --icase -j $PROXY_HSTS_TARGET
		fi
	fi

	iptables -A "$FORWARD" -p udp --dport 53 -j ${PREFIX}dns
	if ${BINDIR}/dnsenabled --log-enabled; then
		$SETCOLOR_FAILURE
		log "WARNING: включаем DNS логирование, отключите его после завершения дебага"
		$SETCOLOR_NORMAL
		echo "${filter['dns_log_level']}" > /proc/net/ipt_dnsmatch/trace_ctl
	fi
	add_custom_rules_dns
}

revisor_hook() {
	return 0
}

add_forward_for_ip() {
	local ext_ip="$1"
	local ext_dev="$2"
	local ip="$3"
	log "Выпускаем $ip в интернет через себя под IP $ext_ip через $ext_dev"
	iptables -A "$FORWARD" -s "$ip" -j ACCEPT
	iptables -A "$FORWARD" -d "$ip" -j ACCEPT
	iptables -t nat -A $POSTROUTING -o "$ext_dev" -s "$ip" -j SNAT --to-source "$ext_ip"
}

revisor() {
	local ext_ip
	local ext_dev
	local ip
	${BINDIR}/dnsenabled --revisor || return 0
	[ "${filter['revisor_route_and_nat']}" = '1' ] || return 0
	if ! ext_ip="$(${BINDIR}/dnsenabled --ext-ip)"; then
		echo "Cant determine default route"
		return 0
	fi
	if ! ext_dev="$(${BINDIR}/dnsenabled --ext-dev)"; then
		echo "Cant determine default route dev"
		return 0
	fi
	if [ -z "$ext_ip" ] || [ -z "$ext_dev" ]; then
		echo "WARNING: Не удалось определить внешний ip и устройство для ревизора"
		return 0
	fi
	for ip in "${revisor_ips[@]}"; do
		add_forward_for_ip "$ext_ip" "$ext_dev" "$ip"
	done
	revisor_hook "$ext_ip" "$ext_dev"
}

filter_input() {
	local WEB_INTERFACE_IP="${admin['web_ip']:-0.0.0.0/0}"
	local WEB_INTERFACE_PORT="${admin['web_port']:-8080}"
	iptables -A "$INPUT" -m state --state ESTABLISHED,RELATED -j ACCEPT
	[ -d /cfg ] && return 0
	iptables -A "$INPUT" -p icmp -j ACCEPT
	iptables -A "$INPUT" -i lo -j ACCEPT
	if [ -n "${admin['ip']:-}" ]; then
		if [ "${filter['forbidden_url_local']:-0}" == '1' ]; then
			iptables -A INPUT -m state --state NEW -p tcp --dport 80 -j ACCEPT
			if ${BINDIR}/dnsenabled; then
				iptables -A INPUT -m state --state NEW -p tcp --dport 443 -j ACCEPT
			fi
		fi
		if [ "$WEB_INTERFACE_PORT" = '80' ] && [ ! -f "$USERDIR/hooks/yes_i_want_web_on_80_port" ]; then
			WEB_INTERFACE_PORT="8080"
		fi
		iptables -A "$INPUT" -m state --state NEW -p tcp -d "${WEB_INTERFACE_IP}" --dport "${WEB_INTERFACE_PORT:-8080}" -m set --match-set reductor_admins src -j ACCEPT
	fi
	iptables -A "$INPUT" -m state --state NEW -m tcp -p tcp --dport "${ssh_port:-22}" -j ACCEPT
	iptables -A "$INPUT" -j REJECT --reject-with icmp-host-prohibited

}

old_reductor_chains_list() {
	local table="$1"
	local regex='^:[^A-Z]+'
	local except_regex='reductor_(maintain|forward|input)'
	iptables-save -t "$table" | egrep "$regex" | egrep -v "$except_regex" | awk '{print $1}' | tr -d ':'
}

firewall_flush_by_numbers() {
	local table="$1"
	local chain="$2"
	while read n _; do
		iptables -t "$table" -D "$chain" "$n"
	done
}

firewall_flush_cr7() {
	local chain
	local table=filter
	# 1 - значит это пришло из restart -> нужно выгрузить модули, либо ipset поменялись
	# по хорошему это разбить на две отдельные части.
	if [ "$MAINTAINANCE_REQUIRED" == '1' ]; then
		${BINDIR}/modules_ctl cleanup_iptables
		${BINDIR}/ipset_ctl cleanup_iptables
		${BINDIR}/modules_ctl unload
	fi
	create_and_flush_chain "$INPUT"
	create_and_flush_chain "$FORWARD"
	# TODO: handle migration to new chains
	# 1. flush всех цепочек кроме основных редукторовских и режима обслуживания:
	for chain in $(old_reductor_chains_list filter FORWARD); do
		log "Чистим $table $chain"
		iptables -t "$table" -F "$chain"
	done
	# 2. почти то же самое делаем внутри самого forward - удаляем ссылки на них.
	chain="FORWARD"
	iptables -t "$table" --line -nL "$chain" | tail -n +3 | egrep -v 'reductor_(maintain|forward)' | tac | firewall_flush_by_numbers "$table" "$chain"
	# 3. уничтожать цепочки не обязательно, пусть болтаются, правил в них нет, модули/ipset они не лочат
}

filter() {
	filter_policy
	filter_input
	add_skip_check_rule
	ip_block
	https
	http
	dns
	ip_port
	revisor
}

mangle() {
	create_and_flush_chain "$PREROUTING" "mangle"
	create_and_flush_chain "$POSTROUTING" "mangle"
}

nat() {
	create_and_flush_chain "$PREROUTING" "nat"
	create_and_flush_chain "$POSTROUTING" "nat"
	[ "${filter['router']}" != '1' ] && return 0
	log "Включается маршрутизация"
	iptables -A "$FORWARD" -j ACCEPT
	{
		iptables -t nat -D $POSTROUTING -j MASQUERADE || true
		iptables -t nat -D $POSTROUTING -j MASQUERADE || true
	} &>/dev/null
	iptables -t nat -I $POSTROUTING -j MASQUERADE
}

raw() {
	local EXT_IP
	iptables -t raw -F PREROUTING
	for chain in mirror_traffic ip_port; do
		create_and_flush_chain "$chain" "raw"
	done
	# shellcheck disable=SC2154
	for link in $(brctl show | egrep -o "$device_regex"); do
		iptables -t raw -I PREROUTING -i "$link" -j mirror_traffic
	done
	if [ "${filter['notrack']:-0}" == '1' ]; then
		if EXT_IP="$(${BINDIR}/dnsenabled --ext-ip)"; then
			if [ -n "$EXT_IP" ]; then
				iptables -t raw -I mirror_traffic -s "$EXT_IP" -j DROP || true
				iptables -t raw -I mirror_traffic -d "$EXT_IP" -j DROP || true
			fi
			iptables -t raw -A mirror_traffic -j NOTRACK
		fi
	fi
	iptables -t raw -A mirror_traffic -m ttl --ttl-eq 1 -j DROP
	if [ "${misc['optimize_routing']}" = '1' ]; then
		iptables -t raw -A mirror_traffic -p tcp -m set --match-set port_http dst -j ACCEPT
		# iptables -t raw -A mirror_traffic -p tcp --dport 443 --tcp-flags SYN SYN -j ACCEPT
		iptables -t raw -A mirror_traffic -p tcp -m set --match-set port_https dst -j ACCEPT
		if [ "${filter['dns']:-0}" = '1' ]; then
			iptables -t raw -A mirror_traffic -p udp --dport 53 -j ACCEPT
		fi
		iptables -t raw -A mirror_traffic -m set --match-set ip_tcp_udp dst -j ACCEPT
		iptables -t raw -A mirror_traffic -m set --match-set ip_block dst -j ACCEPT
		# если дропнуть трафик ревизора, он не попадёт в форвард -> не будет реджектиться.
		# в случае пассивного режима это равносильно пропускам, в активном - неважно.
		if ${BINDIR}/dnsenabled --revisor; then
			local revisor_ip
			for revisor_ip in ${filter['revisor_ip']}; do
				iptables -t raw -A mirror_traffic -s "$revisor_ip" -j ACCEPT
			done
		fi
		iptables -t raw -A mirror_traffic -j DROP
	fi

}

reductor_admins() {
	[ -d /cfg ] && return 0
	local ip
	# ok, no one reads a help text :(
	admin['ip']="${admin['ip']//[,;]/ }"
	ipset flush reductor_admins
	for ip in ${admin['ip']:-}; do
		ipset add reductor_admins "$ip"
	done
}

skip_check() {
	local ip
	if [ -s "$HOOKDIR/skip_check.list" ]; then
		log "Заполняем список пропуска проверок"
		ipset flush skip_check
		while read ip _; do
			if [ "${ip:0:1}" != '#' ]; then
				ipset add skip_check "$ip"
			fi
		done < "$HOOKDIR/skip_check.list"
	fi
}

custom_bindings() {
	${BINDIR}/ipset_ctl custom_bindings_exists || return 0
	local ip ipset
	mkdir -p "$HOOKDIR/custom_ipsets/"
	# shellcheck disable=SC2044
	for ipset in $(find "$HOOKDIR/custom_ipsets/" -type f); do
		ipset flush custom_${ipset##*/}
		while read ip _; do
			ipset add custom_${ipset##*/} "$ip" || true
		done < "$ipset"
	done
}

ipset_prepare() {
	${BINDIR}/ipset_ctl create
	skip_check
	custom_bindings
	reductor_admins
}

# считаем, что IPv6 не используется никем кроме Carbon Reductor.
# живём в FORWARD, никаких специальных цепочек в стиле PL5 не создаём.
ipv6() {
	log "Настраиваем фильтрацию IPv6"
	local traffic_to_filter="-p tcp -m set --match-set port_http dst"
	local TARGET="$TCP_REJECT"
	if [ "${filter['forbidden']}" = '1' ]; then
		local url="${filter['forbidden_url']:-http://helpdesk.carbonsoft.ru/deny}"
		TARGET="FORBIDDEN --forbidden-url $url"
		if [ "${filter['save_domain']:-0}" == '1' ]; then
			TARGET="$TARGET --save-domain"
		fi
		if [ "${filter['forbidden_log']:-0}" == '1' ]; then
			TARGET="$TARGET --log"
		fi
	fi

	echo 0 > /proc/sys/net/ipv6/conf/all/disable_ipv6
	echo 0 > /proc/sys/net/ipv6/conf/default/disable_ipv6
	echo 1 > /proc/sys/net/ipv6/conf/all/forwarding

	# Дублирование загрузки модулей (вдруг врубят ipv6 и перезагрузят фаервол)
	modprobe ip6_tables
	modprobe ip6table_mangle
	modprobe ip6table_filter
	ebtables -t broute -A BROUTING -p IPv6 -j redirect --redirect-target DROP
	ip6tables -P FORWARD DROP
	ip6tables -F FORWARD
	ip6tables -A FORWARD $traffic_to_filter -m reductor --id 0 --comment 'rkn' -j $TARGET
	if ${BINDIR}/dnsenabled; then
		local IP="${filter['dns_ip']:-$CARBONSOFT_BLACKHOLE}"
		ip6tables -A FORWARD -p udp --dport 53 -m dnsmatch --id 0 --comment 'rkn' -j DNSREPLY --dnsreply-ip "$IP"
	fi
	ip6tables -N https &>/dev/null || true
	ip6tables -F https
	for direct in src dst; do
		ip6tables -A FORWARD -p tcp -m set --match-set port_https "$direct" -j https
		for ipset in ip6_https ip6_https_plus; do
			ip6tables -A https -p tcp -m set --match-set "$ipset" "$direct" -j $TCP_REJECT
		done
	done
}

use_hook() {
	hook="$HOOKDIR/${0##*/}"
	if [ -x "$hook" ]; then
		# shellcheck disable=SC1090
		. "$hook" __include
	fi
}

config_hook() {
	if [ "${filter['dns_ip']}" = '127.0.0.1' ]; then
		$SETCOLOR_FAILURE
		log "WARNING: Поскольку не настроена DNS-фильтрация, конфигурируем её автоматически,"
		log "потому что очень беспокоимся о штрафах со стороны Роскомнадзора, а эта функция"
		log "помогает решить имеющуюся проблему с фильтрацией ресурсов с частом меняющимися IP адресами"
		log "'Ненастроенность' определяется тем, что одновременно:"
		log "1) Отключена DNS-фильтрация"
		log "2) IP для редиректа равен 127.0.0.1"
		log "Для настройки запустите menu - настройка алгоритма фильтрации"
		$SETCOLOR_NORMAL
		filter['dns']='1'
		filter['dns_ip']="$CARBONSOFT_BLACKHOLE"
	fi
}

add_link() {
	local table="$1"
	local chain="$2"
	local target="$3"
	if ! iptables -t "$table" -nvL "$chain" | grep -q "$target"; then
		log "Создаём в $table ссылку из $chain в $target"
		iptables -t "$table" -N "$target" 2>/dev/null || true
		# reductor_forward НЕ доджен встать вперёд reductor_maintain, иначе будут пропуски.
		if [ "$target" == 'reductor_forward' ] && iptables -nL FORWARD | grep -q reductor_maintain; then
			iptables -t "$table" -I "$chain" 2 -j "$target"
		else
			iptables -t "$table" -I "$chain" -j "$target"
		fi
	fi
}

start_base_cr7() {
	log "Создаём ссылки в цепочки Reductor"
	add_link "filter" "INPUT" "$INPUT"
	add_link "filter" "FORWARD" "$FORWARD"
	add_link "nat" "PREROUTING" "$PREROUTING"
	add_link "nat" "POSTROUTING" "$POSTROUTING"
	add_link "mangle" "PREROUTING" "$PREROUTING"
	add_link "mangle" "POSTROUTING" "$POSTROUTING"
}

start() {
	log "Конфигурация файрвола"
	# на время перестроения firewall / на время загрузки списков
	if [ "$MAINTAINANCE_REQUIRED" != 0 ] && ! ${BINDIR}/maintain.sh is_locked; then
		${BINDIR}/maintain.sh lock
	fi
	if [ -d /cfg ]; then
		. /usr/local/lib/firewall.shlib
		firewall_flush
	else
		firewall_flush_cr7
		start_base_cr7
	fi
	${BINDIR}/modules_ctl insert
	config_hook
	ipset_prepare
	raw
	mangle
	nat
	filter
	ebtables-restore < "$MAINDIR/contrib/etc/sysconfig/ebtables"
	[ "${filter['ipv6']:-0}" = '1' ] && ipv6
	[ -x "$HOOKDIR/reductor_firewall.sh" ] && ${HOOKDIR}/reductor_firewall.sh start || true
	[ -x "$HOOKDIR/firewall.sh" ] && ${HOOKDIR}/firewall.sh start || true
	# если списки и так все загружены и только перестраиваем firewall
	# то сейчас можно снимать блокировку.
	if [ "$MAINTAINANCE_REQUIRED" = '2' ]; then
		${BINDIR}/maintain.sh unlock
	fi
	log "Конфигурация файрвола завершена"
}

stop() {
	log "Очистка правил фаервола"
	mkdir -p $TMPDIR
	log "Очистка правил фаервола завершена"
}

use_hook
case "${1//[_-]/}" in
	start | stop )
		trap __exit EXIT
		set -eu
		"${1//[_-]/}"
		;;
	: )
		:
		;;
	* )
		echo "$1 not supported, avaible options:"
		grep -A 1000 "case" $0 | grep '[a-z] )' | egrep -o [a-z]+ | sed 's/^/--/g'
		exit 1
		;;
esac
